## Copyright 2022 The IREE Authors
#
# Licensed under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""Defines IREE ARMv8-A benchmarks."""

from typing import List, Sequence

from benchmark_suites.iree import benchmark_presets, module_execution_configs, utils
from e2e_test_framework import unique_ids
from e2e_test_framework.definitions import common_definitions, iree_definitions
from e2e_test_framework.device_specs import device_collections
from e2e_test_framework.models import tflite_models, tf_models, matmul


class Android_ARMv8_A_Benchmarks(object):
    """Benchmarks on ARMv8-A Android devices."""

    MODELS = [
        tflite_models.DEEPLABV3_FP32,
        tflite_models.MOBILEBERT_FP32,
        tf_models.GPT2_117M_1x4_FP32_TF,
        tf_models.GPT2_117M_1x1_FP32_TF,
        tflite_models.MOBILEBERT_INT8,
        tflite_models.VIT_INT8_TFL,
    ]

    MATMULS = [
        matmul.MATMUL_1x256x2048_I8xI4_MLIR,
        matmul.MATMUL_256x256x2048_I8xI4_MLIR,
        matmul.MATMUL_1x256x2048_I8xI8_MLIR,
        matmul.MATMUL_256x256x2048_I8xI8_MLIR,
    ]

    ARMV8_A_CPU_TARGET = iree_definitions.CompileTarget(
        target_architecture=common_definitions.DeviceArchitecture.ARMV8_2_A_GENERIC,
        target_backend=iree_definitions.TargetBackend.LLVM_CPU,
        target_abi=iree_definitions.TargetABI.LINUX_ANDROID29,
    )

    NO_DT_COMPILE_CONFIG = iree_definitions.CompileConfig.build(
        id=unique_ids.IREE_COMPILE_CONFIG_ANDROID_ARMV8_2_A_GENERIC_NO_DT,
        tags=["experimental-flags", "no-dt"],
        compile_targets=[ARMV8_A_CPU_TARGET],
        extra_flags=[
            "--iree-opt-data-tiling=false",
            "--iree-llvmcpu-target-cpu-features=+dotprod",
        ],
    )
    DT_ONLY_COMPILE_CONFIG = iree_definitions.CompileConfig.build(
        id=unique_ids.IREE_COMPILE_CONFIG_ANDROID_ARMV8_2_A_GENERIC_DT_ONLY,
        tags=["experimental-flags", "dt-only"],
        compile_targets=[ARMV8_A_CPU_TARGET],
        extra_flags=[
            "--iree-opt-data-tiling=true",
            "--iree-llvmcpu-enable-ukernels=none",
            "--iree-llvmcpu-target-cpu-features=+dotprod",
        ],
    )
    DT_UK_COMPILE_CONFIG = iree_definitions.CompileConfig.build(
        id=unique_ids.IREE_COMPILE_CONFIG_ANDROID_ARMV8_2_A_GENERIC_DT_UK,
        tags=["default-flags", "dt-uk"],
        compile_targets=[ARMV8_A_CPU_TARGET],
        extra_flags=[
            "--iree-opt-data-tiling=true",
            "--iree-llvmcpu-enable-ukernels=all",
            "--iree-llvmcpu-target-cpu-features=+dotprod",
        ],
    )

    def _build_run_configs(
        self,
        gen_configs: Sequence[iree_definitions.ModuleGenerationConfig],
        exec_configs: Sequence[iree_definitions.ModuleExecutionConfig],
        device_specs: Sequence[common_definitions.DeviceSpec],
        presets: Sequence[str],
    ) -> List[iree_definitions.E2EModelRunConfig]:
        return utils.generate_e2e_model_run_configs(
            module_generation_configs=gen_configs,
            module_execution_configs=exec_configs,
            device_specs=device_specs,
            presets=presets,
        )

    def _generate_model_configs(
        self, device_specs: List[common_definitions.DeviceSpec]
    ) -> List[iree_definitions.E2EModelRunConfig]:
        """Generates IREE compile and run configs for list of models."""
        local_sync_execution_configs = [module_execution_configs.ELF_LOCAL_SYNC_CONFIG]
        local_task_execution_configs = [
            module_execution_configs.get_elf_system_scheduling_local_task_config(
                thread_num
            )
            for thread_num in [1, 2]
        ]
        no_dt_gen_configs = [
            iree_definitions.ModuleGenerationConfig.build(
                compile_config=self.NO_DT_COMPILE_CONFIG,
                imported_model=iree_definitions.ImportedModel.from_model(model),
            )
            for model in self.MODELS
        ]
        dt_only_gen_configs = [
            iree_definitions.ModuleGenerationConfig.build(
                compile_config=self.DT_ONLY_COMPILE_CONFIG,
                imported_model=iree_definitions.ImportedModel.from_model(model),
            )
            for model in self.MODELS
        ]
        dt_uk_gen_configs = [
            iree_definitions.ModuleGenerationConfig.build(
                compile_config=self.DT_UK_COMPILE_CONFIG,
                imported_model=iree_definitions.ImportedModel.from_model(model),
            )
            for model in self.MODELS
        ]
        return self._build_run_configs(
            no_dt_gen_configs + dt_uk_gen_configs,
            local_sync_execution_configs + local_task_execution_configs,
            device_specs,
            [benchmark_presets.ANDROID_CPU],
        ) + self._build_run_configs(
            dt_only_gen_configs,
            local_sync_execution_configs + local_task_execution_configs,
            device_specs,
            [benchmark_presets.ANDROID_CPU_DT_ONLY],
        )

    def _generate_matmul_configs(
        self, device_specs: List[common_definitions.DeviceSpec]
    ) -> List[iree_definitions.E2EModelRunConfig]:
        no_dt_gen_configs = [
            iree_definitions.ModuleGenerationConfig.build(
                compile_config=self.NO_DT_COMPILE_CONFIG,
                imported_model=iree_definitions.ImportedModel.from_model(model),
            )
            for model in self.MATMULS
        ]
        dt_only_gen_configs = [
            iree_definitions.ModuleGenerationConfig.build(
                compile_config=self.DT_ONLY_COMPILE_CONFIG,
                imported_model=iree_definitions.ImportedModel.from_model(model),
            )
            for model in self.MATMULS
        ]
        dt_uk_gen_configs = [
            iree_definitions.ModuleGenerationConfig.build(
                compile_config=self.DT_UK_COMPILE_CONFIG,
                imported_model=iree_definitions.ImportedModel.from_model(model),
            )
            for model in self.MATMULS
        ]
        return self._build_run_configs(
            no_dt_gen_configs + dt_uk_gen_configs,
            [module_execution_configs.ELF_LOCAL_SYNC_CONFIG],
            device_specs,
            [benchmark_presets.ANDROID_CPU],
        ) + self._build_run_configs(
            dt_only_gen_configs,
            [module_execution_configs.ELF_LOCAL_SYNC_CONFIG],
            device_specs,
            [benchmark_presets.ANDROID_CPU_DT_ONLY],
        )

    def generate(
        self,
    ) -> List[iree_definitions.E2EModelRunConfig]:
        """Generates IREE compile and run configs."""
        big_cores_devices = (
            device_collections.DEFAULT_DEVICE_COLLECTION.query_device_specs(
                architecture=common_definitions.DeviceArchitecture.ARMV8_2_A_GENERIC,
                host_environment=common_definitions.HostEnvironment.ANDROID_ARMV8_2_A,
                tags=["big-cores"],
            )
        )

        return self._generate_model_configs(
            big_cores_devices
        ) + self._generate_matmul_configs(big_cores_devices)
